/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.ListIterator;

import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.tagext.BodyContent;
import javax.servlet.jsp.tagext.BodyTagSupport;

import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.XmlPathDiscovery;

import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.UnknownNamespaceException;
import com.topcoder.util.objectfactory.InvalidClassSpecificationException;
import com.topcoder.util.objectfactory.ObjectFactory;
import com.topcoder.util.objectfactory.impl.ConfigManagerSpecificationFactory;
import com.topcoder.util.objectfactory.impl.IllegalReferenceException;
import com.topcoder.util.objectfactory.impl.SpecificationConfigurationException;

/**
 * <p>
 * This tag will render the current node to the root node (or vise-versa) using a path separator and specific styles.
 * </p>
 *
 * <p>
 * The main tag class that application will use.
 * The application can setup the separator and the various styles either by specifying default values in the
 * configuration manager or by specifying attributes on the string. The object factory will be used to discover
 * the node and the path to the root.
 * </p>
 *
 * <p>
 * Note, the node discovery is optional. If the node discovery is not specified, it is assumed a node discovery tag was
 * used on the page to define the current node. The body of the tag will be used to format the various trails
 * (using attributes under page context defined by this class).
 * </p>
 *
 * <p>
 * <b>Thread Safety: </b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 1.0
 *
 */
public class BreadCrumbTrailTag extends BodyTagSupport {
    /**
     * <p>
     * This static constant represents the page scope attribute name that the current node,
     * if found, would be found under.
     * </p>
     *
     * <p>
     * This constant will be used by a NodeDiscover instance to store the current node in the attribute map and
     * by this class to retrieve.
     * </p>
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_CURRENT_NODE = "breadCrumbCurrentNode";

    /**
     * <p>
     * This static constant represents the nested variable name that the current node to format
     * would be found under.
     * </p>
     *
     * <p>
     * This constant will be used by this class to define a variable available to a NodeFormatter to format.
     * </p>
     *
     * <p>
     * Note: this is the page scope attribute name that will be made available as a variable via the
     * BreadCrumbTrailTagExtraInfo class.
     * </p>
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_FORMAT_NODE = "breadCrumbFormatNode";

    /**
     * <p>
     * This static constant represents the nested variable name that the current node's style class that
     * will be used to format.
     * </p>
     *
     * <p>
     * This constant will be used by this class to define a variable available to a NodeFormatter to format.
     * </p>
     *
     * <p>
     * Note: this is the page scope attribute name that will be made available as a variable via the
     * BreadCrumbTrailTagExtraInfo class.
     * </p>
     *
     * @since 2.0
     */
    public static final String BREADCRUMB_FORMAT_STYLE_CLASS = "breadCrumbFormatStyleClass";

    /**
     * <p>
     * This static constant represents a &quot;pathDirection&quot; constant for formatting the path
     * from the root (on the left) to the current node (on the right).
     * </p>
     *
     * <p>
     * This variable is referenced in doStartTag(), doInitBody(), doAfterBody() methods but is public for
     * a convenience to strategy implementations.
     * </p>
     *
     * <p>
     * Note, this constant is the default value for &quot;pathDirection&quot; variable.
     * </p>
     *
     * @since 1.0
     */
    public static final int ROOT_TO_CURRENT = 0;

    /**
     * <p>
     * This static constant represents a &quot;pathDirection&quot;constant for formatting the path
     * from the current node (on the left) to the root node (on the right).
     * </p>
     *
     * <p>
     * This variable is referenced in doStartTag(), doInitBody(), doAfterBody() methods but is public for
     * a convenience to strategy implementations.
     * </p>
     *
     * @since 1.0
     */
    public static final int CURRENT_TO_ROOT = 1;

    /**
     * <p>
     * This static constant represents the namespace used by this class.
     * The configuration information is read under this namespace via Configuration Manager Component.
     * </p>
     *
     * @since 2.0
     */
    private static final String NAMESPACE = "com.topcoder.web.ui.tag.breadcrumbtrail";

    /**
     * <p>
     * This static constant represents the pathSeparator property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String PATH_SEPARATOR = "pathSeparator";

    /**
     * <p>
     * This static constant represents the pathSeparatorStyle property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String PATH_SEPARATOR_STYLE = "pathSeparatorStyle";

    /**
     * <p>
     * This static constant represents the nodeStyle property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String NODE_STYLE = "nodeStyle";

    /**
     * <p>
     * This static constant represents the rootNodeStyle property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String ROOT_NODE_STYLE = "rootNodeStyle";

    /**
     * <p>
     * This static constant represents the currentNodeStyle property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String CURRENT_NODE_STYLE = "currentNodeStyle";

    /**
     * <p>
     * This static constant represents the mouseOverStyle property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String MOUSEOVER_STYLE = "mouseOverStyle";

    /**
     * <p>
     * This static constant represents the pathDirection property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String PATH_DIRECTION = "pathDirection";

    /**
     * <p>
     * This static constant represents the silentErrors property name.
     * </p>
     *
     * @since 2.0
     */
    private static final String SILENT_ERRORS = "silentErrors";

    /**
     * <p>
     * This static constant represents the node style class name.
     * </p>
     *
     * @since 2.0
     */
    private static final String NODE_STYLE_CLASS_NAME = "node-style";

    /**
     * <p>
     * This static constant represents the current node style class name.
     * </p>
     *
     * @since 2.0
     */
    private static final String CURRENT_NODE_STYLE_CLASS_NAME = "current-node-style";

    /**
     * <p>
     * This static constant represents the root node style class name.
     * </p>
     *
     * @since 2.0
     */
    private static final String ROOT_NODE_STYLE_CLASS_NAME = "root-node-style";

    /**
     * <p>
     * This static constant represents the path separator style class name.
     * </p>
     *
     * @since 2.0
     */
    private static final String PATH_SEPARATOR_STYLE_CLASS_NAME = "path-separator-style";

    /**
     * <p>
     * Represents the separator that will be drawn between nodes in a path.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and will never be null but can be
     * an empty (trim'd) string. The default value is &quot;:&quot;.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody
     * methods.
     * </p>
     *
     * @since 1.0
     */
    private String pathSeparator = ":";

    /**
     * <p>
     * Represents the separator CSS style that will be used to draw the separator.
     * This is the actual CSS style that will be rendered to the underlying text.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (i.e. use the style that is currently in context).
     * The default value is null.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody methods.
     * </p>
     *
     * @since 1.0
     */
    private String pathSeparatorStyle;

    /**
     * <p>
     * Represents the node (non-root and non-current) CSS style that will be used to draw the
     * node's title. This is the actual CSS style that will be rendered to the underlying link.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (i.e. use the style that is currently in context).
     * The default value is null.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody methods.
     * </p>
     *
     * @since 1.0
     */
    private String nodeStyle;

    /**
     * <p>
     * Represents the current (only) node CSS style that will be used to draw the node's title.
     * This is the actual CSS style that will be rendered to the underlying link.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (i.e. use the style that is currently in context).
     * The default value is null.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody methods.
     * </p>
     *
     * @since 1.0
     */
    private String currentNodeStyle;

    /**
     * <p>
     * Represents the root (only) node CSS style that will be used to draw the node's title.
     * This is the actual CSS style that will be rendered to the underlying link.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (i.e. use the style that is currently in context).
     * The default value is null.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody methods.
     * </p>
     *
     * @since 1.0
     */
    private String rootNodeStyle;

    /**
     * <p>
     * Represents the mouse over node CSS style that will be used to draw the node's title.
     * This is the actual CSS style that will be rendered to the underlying link.
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can be null or an empty (trim'd) string
     * to specify no specific style (i.e. use the style that is currently in context).
     * The default value is null.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doAfterBody methods.
     * </p>
     *
     * @since 1.0
     */
    private String mouseOverStyle;

    /**
     * <p>
     * Represents whether any exceptions should be silently ignore or not (true to silently ignore,
     * false to let exception pass through to Jsp engine).
     * </p>
     *
     * <p>
     * This variable is set by the constructor and is mutable.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doStartTag, doAfterBody and
     * doEndTag methods.
     * </p>
     *
     * @since 1.0
     */
    private boolean silentErrors = false;

    /**
     * <p>
     * Represents the direction of the path.
     * The current allowable values are :
     * <ol>
     * <li>ROOT_TO_CURRENT - specifying root [on the left] to current [on the right]</li>
     * <li>CURRENT_TO_ROOT - specifying current [on the left] to root [on the right]</li>
     * </ol>
     * </p>
     *
     * <p>
     * This variable is set by the constructor, is mutable and can only be the above values.
     * The default value is ROOT_TO_CURRENT.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doStartTag methods.
     * </p>
     *
     * @since 1.0
     */
    private int pathDirection = ROOT_TO_CURRENT;

    /**
     * <p>
     * Represents the a path override for this specific tag.
     * </p>
     *
     * <p>
     * This variable is set initially to null and is mutable. If null, there is no path override available
     * and normal path discovery should be used. Of non-null, a XmlPathDiscovery class will be constructed with
     * this path and used to return the path specified in this variable.
     * </p>
     *
     * <p>
     * Note, this string contains the direct xml string, please see XmlPathDiscovery class for details on it's
     * formatting.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doStartTag methods.
     * </p>
     *
     * @since 2.0
     */
    private String pathOverride;

    /**
     * <p>
     * Represents title that will be used if the node is being discovered by this tag (i.e. no other tag has already
     * discovered the node and set the BREADCRUMB_CURRENT_NODE attribute) - otherwise it is ignored.
     * </p>
     *
     * <p>
     * This variable is set initially to null, is mutable and never will be an empty string but can be null.
     * A null indicates that the title should default to the title specified in the PathDiscovery data source.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods and the doStartTag methods.
     * </p>
     *
     * @since 2.0
     */
    private String title;

    /**
     * <p>
     * Represents the iterator over the current node path.
     * </p>
     *
     * <p>
     * This iterator will initially be null, be mutable by the doStartTag and will be non-null during
     * the body processing. The iterator will be created in the doStartTag from the path discovered by
     * the PathDiscovery implementation.
     * </p>
     *
     * <p>
     * The node formatter (or body if no node formatter specified) will then be called, for each node
     * in this iterator.  This variable will only be non-null if a node formatter is not specified (i.e. that
     * we need the body to format each node).
     * </p>
     *
     * @since 2.0
     */
    private ListIterator iterator;

    /**
     * <p>
     * Constructs the BreadCrumbTrailTag using the configuration manager to set default values.
     * </p>
     *
     * <p>
     * Note, If a configuration exception occurs on any step, the exception will be ignored and default value is used.
     * </p>
     *
     * @since 2.0
     */
    public BreadCrumbTrailTag() {
        // config pathSepartor via config manager, default value is ":"
        this.pathSeparator = getConfigString(PATH_SEPARATOR, ":", false);

        // config pathSeparatorStye, nodeStyle, currentNodeStyle, rootNodeStyle and mouseOverStyle via config manager.
        // default value is null, if empty config value is found, then null will be returned instead.
        this.pathSeparatorStyle = getConfigString(PATH_SEPARATOR_STYLE, null, true);
        this.nodeStyle = getConfigString(NODE_STYLE, null, true);
        this.currentNodeStyle = getConfigString(CURRENT_NODE_STYLE, null, true);
        this.rootNodeStyle = getConfigString(ROOT_NODE_STYLE, null, true);
        this.mouseOverStyle = getConfigString(MOUSEOVER_STYLE, null, true);

        // config silentErrors via config manager, default value is false
        this.silentErrors = Boolean.valueOf(getConfigString(SILENT_ERRORS, "false", false)).booleanValue();

        // config pathDirection via config manager, default value is ROOT_TO_CURRENT.
        // if invalid config value is found, then ROOT_TO_CURRENT will be used instead.
        try {
            int pathDirectionValue = Integer.parseInt(getConfigString(PATH_DIRECTION, "0", false));
            if (pathDirectionValue == ROOT_TO_CURRENT || pathDirectionValue == CURRENT_TO_ROOT) {
                this.pathDirection = pathDirectionValue;
            }
        } catch (NumberFormatException nfe) {
            // ignore, the default value ROOT_TO_CURRENT is used.
        }

    }

    /**
     * <p>
     * Returns value of pathDirection property.
     * </p>
     *
     * @return The path direction - either ROOT_TO_CURRENT or CURRENT_TO_ROOT
     * @see BreadCrumbTrailTag#pathDirection
     * @since 1.0
     */
    public int getPathDirection() {
        return pathDirection;
    }

    /**
     * <p>
     * Sets value of pathDirection property.
     * </p>
     *
     * @param pathDirection The new path direction
     *
     * @throws IllegalArgumentException if pathDirection is not ROOT_TO_CURRENT or CURRENT_TO_ROOT
     * @see BreadCrumbTrailTag#pathDirection
     * @since 1.0
     */
    public void setPathDirection(int pathDirection) {
        if (pathDirection != ROOT_TO_CURRENT && pathDirection != CURRENT_TO_ROOT) {
            throw new IllegalArgumentException(
                "The value of pathDirection should be ROOT_TO_CURRENT or CURRENT_TO_ROOT.");
        }

        this.pathDirection = pathDirection;
    }

    /**
     * <p>
     * Returns value of pathSeparator property.
     * </p>
     *
     * @return A non-null, possibly empty (trim'd) string representing the path separator
     * @see BreadCrumbTrailTag#pathSeparator
     * @since 1.0
     */
    public String getPathSeparator() {
        return pathSeparator;
    }

    /**
     * <p>
     * Sets value of pathSeparator property.
     * </p>
     *
     * @param pathSeparator A non-null, possibly empty (trim'd) string representing the path separator
     * @throws IllegalArgumentException if the pathSeparator is null
     * @see BreadCrumbTrailTag#pathSeparator
     * @since 1.0
     */
    public void setPathSeparator(String pathSeparator) {
        Util.checkNull(pathSeparator, "pathSeparator");

        this.pathSeparator = pathSeparator;
    }

    /**
     * <p>
     * Returns value of pathSeparatorStyle Property.
     * </p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the path separator style
     * @see BreadCrumbTrailTag#pathSeparatorStyle
     * @since 1.0
     */
    public String getPathSeparatorStyle() {
        return pathSeparatorStyle;
    }

    /**
     * <p>
     * Sets value of pathSeparatorStyle property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be used instead.
     * </p>
     *
     * @param pathSeparatorStyle A possibly null, possibly empty (trim'd) string representing the path separator
     * style (empty strings will be translated to a null)
     * @see BreadCrumbTrailTag#pathSeparatorStyle
     * @since 2.0
     */
    public void setPathSeparatorStyle(String pathSeparatorStyle) {
        this.pathSeparatorStyle = (Util.isEmptyString(pathSeparatorStyle) ? null : pathSeparatorStyle);
    }

    /**
     * <p>
     * Returns value of nodeStyle property.
     * </p>
     *
     * @return A possibly null, never empty (trim'd) string representing the node style
     * @see BreadCrumbTrailTag#nodeStyle
     * @since 1.0
     */
    public String getNodeStyle() {
        return nodeStyle;
    }

    /**
     * <p>
     * Sets value of nodeStyle property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param nodeStyle A possibly null, possibly empty (trim'd) string representing the node style
     * (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#nodeStyle
     * @since 1.0
     */
    public void setNodeStyle(String nodeStyle) {
        this.nodeStyle = (Util.isEmptyString(nodeStyle) ? null : nodeStyle);
    }

    /**
     * <p>
     * Returns value of currentNodeStyle property.
     * </p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the current node style
     * @see BreadCrumbTrailTag#currentNodeStyle
     * @since 1.0
     */
    public String getCurrentNodeStyle() {
        return currentNodeStyle;
    }

    /**
     * <p>
     * Sets value of currentNodeStyle property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param currentNodeStyle A possibly null, possibly empty (trim'd) string representing the current
     * node style (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#currentNodeStyle
     * @since 1.0
     */
    public void setCurrentNodeStyle(String currentNodeStyle) {
        this.currentNodeStyle = (Util.isEmptyString(currentNodeStyle) ? null : currentNodeStyle);
    }

    /**
     * <p>
     * Returns value of rootNodeStyle property.
     * </p>
     *
     *
     * @return A possibly null, never empty (trim'd) string representing the root node style
     * @see BreadCrumbTrailTag#rootNodeStyle
     * @since 1.0
     */
    public String getRootNodeStyle() {
        return rootNodeStyle;
    }

    /**
     * <p>
     * Sets value of rootNodeStyle property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param rootNodeStyle A possibly null, possibly empty (trim'd) string representing the root node style
     * (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#rootNodeStyle
     * @since 1.0
     */
    public void setRootNodeStyle(String rootNodeStyle) {
        this.rootNodeStyle = (Util.isEmptyString(rootNodeStyle) ? null : rootNodeStyle);
    }

    /**
     * <p>
     * Returns value of mouseOverStyle property.
     * </p>
     *
     * @return A possibly null, never empty (trim'd) string representing the mouse over style
     * @see BreadCrumbTrailTag#mouseOverStyle
     * @since 1.0
     */
    public String getMouseOverStyle() {
        return mouseOverStyle;
    }

    /**
     * <p>
     * Sets value of mouseOverStyle property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param mouseOverStyle A possibly null, possibly empty (trim'd) string representing the mouse
     * over style (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#mouseOverStyle
     * @since 1.0
     */
    public void setMouseOverStyle(String mouseOverStyle) {
        this.mouseOverStyle = (Util.isEmptyString(mouseOverStyle) ? null : mouseOverStyle);
    }

    /**
     * <p>
     * Returns value of pathOverride property.
     * </p>
     *
     * @return A possibly null, never empty (trim'd) string representing the path override
     * @see BreadCrumbTrailTag#pathOverride
     * @since 2.0
     */
    public String getPathOverride() {
        return pathOverride;
    }

    /**
     * <p>
     * Sets value of pathOverride property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param pathOverride A possibly null, possibly empty (trim'd) string representing the path override
     * (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#pathOverride
     * @since 2.0
     */
    public void setPathOverride(String pathOverride) {
        this.pathOverride = (Util.isEmptyString(pathOverride) ? null : pathOverride);
    }

    /**
     * <p>
     * Returns value of silentErrors property.
     * </p>
     *
     * @return True if exceptions will be silently ignored, false if they should be thrown to the Jsp Engine
     * @see BreadCrumbTrailTag#silentErrors
     * @since 1.0
     */
    public boolean getSilentErrors() {
        return silentErrors;
    }

    /**
     * <p>
     * Sets value of silentErrors property.
     * </p>
     *
     * @param silentErrors True if exceptions will be silently ignored, false if they should be thrown to the Jsp Engine
     * @see BreadCrumbTrailTag#silentErrors
     * @since 1.0
     */
    public void setSilentErrors(boolean silentErrors) {
        this.silentErrors = silentErrors;
    }

    /**
     * <p>
     * Returns value of title property.
     * </p>
     *
     * @return A possibly null, never empty (trim'd) string representing the current node title
     * @see BreadCrumbTrailTag#title
     * @since 2.0
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>
     * Sets value of title property.
     * </p>
     *
     * <p>
     * Note, if the argument is an empty (trim'd) string, null will be instead.
     * </p>
     *
     * @param title A possibly null, possibly empty (trim'd) string representing the title of the current node
     * (an empty string will be translated to a null)
     * @see BreadCrumbTrailTag#title
     * @since 2.0
     */
    public void setTitle(String title) {
        this.title = (Util.isEmptyString(title) ? null : title);
    }

    /**
     * <p>
     * Initializes the environment for further processing, including the current node discovery and the current path
     * discovery.
     * </p>
     *
     * <p>
     * If any style is configed then the style settings will be printed out to the specified writer.
     * </p>
     *
     * <p>
     * Note, if silentErrors is false, then any exception will cause BreadCrumbTrailTag to be thrown, otherwise,
     * SKIP_BODY will be returned to silently ignore exception and stop further tag processing.
     * </p>
     *
     * @return An integer representing what processing to do after this method executes(it's defined by tag life cycle)
     *
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is false
     * @since 1.0
     */
    public int doStartTag() throws BreadCrumbException {
        // discovery the current node
        BreadCrumbNode node = null;
        // try to get the current node from page context attribute
        Object obj = pageContext.getAttribute(BREADCRUMB_CURRENT_NODE);
        if (obj instanceof BreadCrumbNode) {
            node = (BreadCrumbNode) obj;
        } else {
            // if not found in page context attribute, then construct a NodeDiscovery instance from
            // object factory to return the current node.
            NodeDiscovery nodeDiscovery = createNodeDiscovery();
            // fail to construct a NodeDiscovery instance and no exception be thrown, then silentErrors
            // must be set true, in this case, SKIP_BODY will be returned to skip further processing.
            if (nodeDiscovery == null) {
                return SKIP_BODY;
            }
            if (nodeDiscovery instanceof AbstractNodeDiscovery && getTitle() != null) {
                ((AbstractNodeDiscovery) nodeDiscovery).setTitle(getTitle());
            }
            node = nodeDiscovery.getCurrentNode(pageContext);
            // current node is not found, then ignore or throw an exception depends the silentErrors property.
            if (node == null) {
                if (silentErrors) {
                    return SKIP_BODY;
                } else {
                    throw new BreadCrumbException("Fail to discovery the current bread crumb node.");
                }
            }
        }

        PathDiscovery pathDiscovery = getPathDiscovery();

        // fail to construct a PathDiscovery instance and no exception be thrown, then silentErrors
        // must be set true, in this case, SKIP_BODY will be returned to skip further processing.
        if (pathDiscovery == null) {
            return SKIP_BODY;
        }

        BreadCrumbNode[] nodes = null;

        try {
            nodes = pathDiscovery.getPath(node, pageContext);
        } catch (BreadCrumbException e) {
            return handleBreadCrumbException(e);
        }

        // no path can be found, SKIP_BODY will be returned to skip further processing.
        if (nodes == null || nodes.length == 0) {
            return SKIP_BODY;
        }

        List nodesList = Arrays.asList(nodes);

        // set the initial position for iterator instance.
        // for position issue while accessing via next() or previous(), please refer to
        // method documentation for List#listIterator(int).
        if (pathDirection == ROOT_TO_CURRENT) {
            iterator = nodesList.listIterator(0);
        } else {
            iterator = nodesList.listIterator(nodesList.size());
        }

        // style setting will be written out if any style is configed
        if (!isAllStyleVariablesNull()) {
            JspWriter writer = pageContext.getOut();
            try {
                writer.write("<span>");
                printStyleSettings(writer);
            } catch (IOException e) {
                return handleIOException(e);
            }
        }

        return EVAL_BODY_BUFFERED;
    }

    /**
     * <p>
     * Get's the next item from the iterator and set's the attribute information(including BREADCRUMB_FORMAT_NODE
     * and BREADCRUMB_FORMAT_STYLE_CLASS attributes) for what needs to be formatted.
     * </p>
     *
     * @since 2.0
     */
    public void doInitBody() {
        // this method is reused by doInitBody() and doAfterBody() methods.
        setCurrentFormatNode();
    }

    /**
     * <p>
     * Get's the body content text and write's it out and starts again.
     * </p>
     *
     * <p>
     * In this method, if next node exists, BREADCRUMB_FORMAT_NODE and BREADCRUMB_FORMAT_STYLE_CLASS attributes will be
     * set to the next node and the tag body will be evaluated again.
     * </p>
     *
     * <p>
     * The sequence of node to be formatted depends on the pathDirection property.
     * <ol>
     * <li>If ROOT_TO_CURRENT is being used, the next node to be formatted is the next element in the iterator.</li>
     * <li>If CURRENT_TO_ROOT is being used, the next node to be formatted is the previous element in the iterator.</li>
     * </ol>
     * </p>
     *
     * <p>
     * Note, if silentErrors is false, then any exception will cause BreadCrumbTrailTag to be thrown, otherwise,
     * SKIP_BODY will be returned to silently ignore exception and stop further tag processing.
     * </p>
     *
     * @return Either SKIP_BODY to stop processing the body or EVAL_BODY_AGAIN to process the next item
     *
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is false
     *
     * @since 2.0
     */
    public int doAfterBody() throws BreadCrumbException {

        JspWriter writer = getPreviousOut();

        BodyContent bodyContent = getBodyContent();
        String content = bodyContent.getString();

        // write the evaluated body content in case that it's not null or empty.
        if (content != null && content.trim().length() != 0) {
            try {
                writer.write(content);
                bodyContent.clear();
            } catch (IOException e) {
                return handleIOException(e);
            }
        }

        // the next node exists or not depends on pathDirection.
        boolean loopAgain = false;
        if (pathDirection == ROOT_TO_CURRENT) {
            loopAgain = iterator.hasNext();
        } else {
            loopAgain = iterator.hasPrevious();
        }

        if (loopAgain) {
            try {
                printSeparator(writer);
                setCurrentFormatNode();
            } catch (IOException e) {
                return handleIOException(e);
            }
            return EVAL_BODY_AGAIN;
        } else {
            return SKIP_BODY;
        }
    }

    /**
     * <p>
     * Prints the style settings out to the specified writer if any style is specified.
     * </p>
     *
     * @return Always EVAL_PAGE to continue evaluating the page
     *
     * @throws BreadCrumbException if creating of bread crumb trail failed and silentErrors is false
     *
     * @since 2.0
     */
    public int doEndTag() throws BreadCrumbException {
        if (!isAllStyleVariablesNull()) {
            JspWriter writer = pageContext.getOut();
            try {
                writer.write("</span>");
            } catch (IOException e) {
                if (!silentErrors) {
                    throw new BreadCrumbException("IOException occurs, caused by " + e.getMessage(), e);
                }
            }
        }

        return EVAL_PAGE;
    }

    /**
     * <p>
     * Prints the style settings out to the specified writer.
     * </p>
     *
     * @param out The non-null JspWriter to write the style settings to
     *
     * @throws IllegalArgumentException if the out argument is null
     * @throws IOException if an I/O error occurs writing out the style
     *
     * @since 1.0
     */
    private void printStyleSettings(JspWriter out) throws IOException {
        Util.checkNull(out, "out");

        out.write("<style>\n");

        if (getPathSeparatorStyle() != null) {
            printStyle(out, "." + PATH_SEPARATOR_STYLE_CLASS_NAME, getPathSeparatorStyle());
        }

        if (getNodeStyle() != null) {
            printStyle(out, "a." + NODE_STYLE_CLASS_NAME + ":link", getNodeStyle());
        }

        if (getCurrentNodeStyle() != null) {
            printStyle(out, "a." + CURRENT_NODE_STYLE_CLASS_NAME + ":link", getCurrentNodeStyle());
        }

        if (getRootNodeStyle() != null) {
            printStyle(out, "a." + ROOT_NODE_STYLE_CLASS_NAME + ":link", getRootNodeStyle());
        }

        if (getMouseOverStyle() != null) {
            printStyle(out, "a." + NODE_STYLE_CLASS_NAME + ":hover", getMouseOverStyle());

            if (getCurrentNodeStyle() != null) {
                printStyle(out, "a." + CURRENT_NODE_STYLE_CLASS_NAME + ":hover", getMouseOverStyle());
            }

            if (getRootNodeStyle() != null) {
                printStyle(out, "a." + ROOT_NODE_STYLE_CLASS_NAME + ":hover", getMouseOverStyle());
            }
        }

        out.write("</style>\n");

    }

    /**
     * <p>
     * Prints the specified style out to the specified writer.
     * </p>
     *
     * @param out The non-null JspWriter to write the style settings to
     * @param style The non-null, non-empty (trim'd) string representing the style to write
     * @param value The non-null, non-empty (trim'd) string representing the value of the style to write
     *
     * @throws IllegalArgumentException if any argument is null or any string argument is an empty (trim'd) string
     * @throws IOException if an I/O error occurs writing out the style
     *
     * @since 1.0
     */
    private void printStyle(JspWriter out, String style, String value) throws IOException {
        Util.checkNull(out, "out");
        Util.checkString(style, "style");
        Util.checkString(value, "value");

        out.write(style);
        out.write(" {");
        out.write(value);
        out.write("}\n");
    }

    /**
     * <p>
     * Prints the separator (with the separator style) to the specified writer.
     * </p>
     *
     * @param out The non-null JspWriter to write the style settings to
     *
     * @throws IllegalArgumentException if the out is null
     * @throws IOException if an I/O error occurs writing out the separator(with the separator style)
     *
     * @since 1.0
     */
    private void printSeparator(JspWriter out) throws IOException {
        Util.checkNull(out, "out");

        // the pathSeparator will be embedded in a span style tag when the pathSeparatorStyle is not null,
        // otherwise only pathSeparator will be written out.
        if (pathSeparatorStyle != null) {
            out.write("<span class=\"" + PATH_SEPARATOR_STYLE_CLASS_NAME + "\">");
            out.write(getPathSeparator());
            out.write("</span>\n");
        } else {
            out.write(getPathSeparator());
        }
    }

    /**
     * <p>
     * Sets the BREADCRUMB_FORMAT_NODE and BREADCRUMB_FORMAT_STYLE_CLASS attributes in page context with
     * the corresponding values.
     * </p>
     *
     * <p>
     * BREADCRUMB_FORMAT_NODE attribute will store the current node to be formatted.
     * BREADCRUMB_FORMAT_STYLE_CLASS will store the style class to be applied to the current node to be formatted.
     * </p>
     *
     * <p>
     * Note, pathDirection and iterator field is used to choose the correct node and style.
     * If multiple style classes apply to the current node, then the priority is :
     * <b>root-node-style &gt; current-node-style &gt; node-style</b>
     * For example, if there's only one node in the current path, then the style applies to the only node will be
     * root-node-style.
     * </p>
     *
     * @since 2.0
     */
    private void setCurrentFormatNode() {
        BreadCrumbNode toFormatNode = null;
        String style = null;

        if (pathDirection == ROOT_TO_CURRENT) {
            toFormatNode = (BreadCrumbNode) iterator.next();

            // get the style to be applied to the current node to be formatted
            // it is put after calling next() method so the nextIndex() will return the position of next node
            // please refer to the documentation of ListIterator
            // note, in priority, root over current over node
            if (iterator.nextIndex() == 1) {
                style = ROOT_NODE_STYLE_CLASS_NAME;
            } else if (!iterator.hasNext()) {
                style = CURRENT_NODE_STYLE_CLASS_NAME;
            } else {
                style = NODE_STYLE_CLASS_NAME;
            }
        } else {
            // get the style to be applied to the current node to be formatted
            // it is put before calling previous() method so the previousIndex() will return the position of
            // current node, please refer to the documentation of ListIterator
            // note, in priority, root over current over node
            if (iterator.previousIndex() == 0) {
                style = ROOT_NODE_STYLE_CLASS_NAME;
            } else if (!iterator.hasNext()) {
                style = CURRENT_NODE_STYLE_CLASS_NAME;
            } else {
                style = NODE_STYLE_CLASS_NAME;
            }

            toFormatNode = (BreadCrumbNode) iterator.previous();
        }

        pageContext.setAttribute(BREADCRUMB_FORMAT_NODE, toFormatNode);
        pageContext.setAttribute(BREADCRUMB_FORMAT_STYLE_CLASS, style);
    }

    /**
     * <p>
     * Returns true when none style is specified, otherwise return false.
     * </p>
     *
     * @return true when none style is specified, otherwise return false.
     *
     * @since 2.0
     */
    private boolean isAllStyleVariablesNull() {
        // there are five style properties in all.
        if (this.mouseOverStyle != null) {
            return false;
        }

        if (this.nodeStyle != null) {
            return false;
        }

        if (this.currentNodeStyle != null) {
            return false;
        }

        if (this.pathSeparatorStyle != null) {
            return false;
        }

        if (this.rootNodeStyle != null) {
            return false;
        }

        return true;
    }

    /**
     * <p>
     * Return the value of the property.
     * </p>
     *
     * <p>
     * Note, if the given convertEmptyString is true, then when the config value is empty, null will be returned
     * instead.
     * </p>
     *
     * @param propertyName the name of property.
     * @param defaultValue if the config value is not found, then this value will be returned instead.
     * @param convertEmptyString flag to decide whether to convert the empty string to null or not.
     *
     * @return the value of the property.
     *
     * @since 2.0
     */
    private String getConfigString(String propertyName, String defaultValue, boolean convertEmptyString) {
        String property = null;

        try {
            property = ConfigManager.getInstance().getString(NAMESPACE, propertyName);
        } catch (UnknownNamespaceException e1) {
            return defaultValue;
        }

        // property is missing
        if ((property == null)) {
            return defaultValue;
        }

        // empty config value and converEmptyString is true, then return null.
        if (convertEmptyString && property.trim().length() == 0) {
            return null;
        }

        return property;
    }

    /**
     * <p>
     * Constructs a NodeDiscovery instance via object factory.
     * </p>
     *
     * <p>
     * The namespace to be used is the fully qualified name of BreadCrumbTrailTag.
     * </p>
     *
     * <p>
     * Note, if silentErrors is false, then any exception will cause BreadCrumbTrailTag to be thrown, otherwise,
     * null will be returned to silently ignore exception.
     * </p>
     *
     * @return a new NodeDiscovery instance dynamically created via object factory.
     *
     * @throws BreadCrumbException when failed to create NodeDiscovery instance via object factory
     * @since 2.0
     */
    private NodeDiscovery createNodeDiscovery() throws BreadCrumbException {
        try {
            ObjectFactory factory = new ObjectFactory(new ConfigManagerSpecificationFactory(
                    BreadCrumbTrailTag.class.getName()));
            // in the config file, the property name should be the fully qualified name of NodeDiscovery interface
            // with dot be replaced with slash
            return (NodeDiscovery) factory.createObject(NodeDiscovery.class);
        } catch (SpecificationConfigurationException e) {
            if (!silentErrors) {
                throw new BreadCrumbException(
                        "SpecificationConfigurationException occurs while constructing NodeDiscovery instance, "
                        + "caused by " + e.getMessage(), e);
            }
        } catch (IllegalReferenceException e) {
            if (!silentErrors) {
                throw new BreadCrumbException("IllegalReferenceException occurs while constructing NodeDiscovery "
                        + "instance, caused by " + e.getMessage(), e);
            }
        } catch (InvalidClassSpecificationException e) {
            if (!silentErrors) {
                throw new BreadCrumbException("InvalidClassSpecificationException occurs while constructing "
                        + "NodeDiscovery instance, caused by " + e.getMessage(), e);
            }
        }
        return null;
    }

    /**
     * <p>
     * Constructs a PathDiscovery instance via object factory.
     * </p>
     *
     * <p>
     * The namespace to be used is the fully qualified name of BreadCrumbTrailTag.
     * </p>
     *
     * <p>
     * Note, if silentErrors is false, then any exception will cause BreadCrumbTrailTag to be thrown, otherwise, null
     * will be returned to silently ignore exception.
     * </p>
     *
     * @return a new PathDiscovery instance dynamically created via object factory.
     * @throws BreadCrumbException when failed to create PathDiscovery instance via object factory
     * @since 2.0
     */
    private PathDiscovery createPathDiscovery() throws BreadCrumbException {
        try {
            ObjectFactory factory = new ObjectFactory(new ConfigManagerSpecificationFactory(
                    BreadCrumbTrailTag.class.getName()));
            // in the config file, the property name should be the fully qualified name of PathDiscovery interface
            //with dot be replaced with slash
            return (PathDiscovery) factory.createObject(PathDiscovery.class);
        } catch (SpecificationConfigurationException e) {
            if (!silentErrors) {
                throw new BreadCrumbException(
                        "SpecificationConfigurationException occurs while constructing PathDiscovery instance, "
                        + "caused by " + e.getMessage(), e);
            }
        } catch (IllegalReferenceException e) {
            if (!silentErrors) {
                throw new BreadCrumbException("IllegalReferenceException occurs while constructing PathDiscovery "
                        + "instance, caused by " + e.getMessage(), e);
            }
        } catch (InvalidClassSpecificationException e) {
            if (!silentErrors) {
                throw new BreadCrumbException("InvalidClassSpecificationException occurs while constructing "
                        + "PathDiscovery instance, caused by " + e.getMessage(), e);
            }
        }
        return null;
    }

    /**
     * <p>
     * This method handles the situation when a BreadCrumbException occurs.
     * </p>
     *
     * <p>
     * If silentErrors is true, then return SKIP_BODY to skip body evaluation, otherwise rethrow the
     * BreadCrumbException.
     * </p>
     *
     * @param e BreadCrumbException exception thrown in tag processing
     * @return SKIP_BODY to skip body evaluation
     *
     * @throws BreadCrumbException when silentErrors is false
     */
    private int handleBreadCrumbException(BreadCrumbException e) throws BreadCrumbException {
        if (silentErrors) {
            return SKIP_BODY;
        } else {
            throw e;
        }
    }

    /**
     * <p>
     * This method handles the situation when an IOException occurs.
     * </p>
     *
     * <p>
     * If silentErrors is true, then return SKIP_BODY to skip body evaluation, otherwise wrap the
     * IOException and throw BreadCrumbException.
     * </p>
     *
     * @param e IOException exception thrown in tag processing
     * @return SKIP_BODY to skip body evaluation
     *
     * @throws BreadCrumbException when silentErrors is false
     */
    private int handleIOException(IOException e) throws BreadCrumbException {
        if (silentErrors) {
            return SKIP_BODY;
        } else {
            throw new BreadCrumbException("IOException occurs, caused by " + e.getMessage(), e);
        }
    }

    /**
     * <p>
     * Returns PathDiscovery instance to discovery the current path.
     * </p>
     *
     * <p>
     * The PathDiscovery instance can be XmlPathDiscovery if pathOverride is specified or is created via object
     * factory.
     * </p>
     *
     * @return PathDiscovery instance to discovery the current path.
     *
     * @throws BreadCrumbException if creating of PathDiscovery instance failed and silentErrors is false
     */
    private PathDiscovery getPathDiscovery() throws BreadCrumbException {
        // discovery the current path
        PathDiscovery pathDiscovery = null;
        if (pathOverride != null) {
            pathDiscovery = new XmlPathDiscovery(getPathOverride(), false);
        } else {
            // if pathOverride is not specified, create a PathDiscovery instance via object factory.
            pathDiscovery = createPathDiscovery();
        }
        return pathDiscovery;
    }
}
