/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodediscovery;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;

import com.topcoder.web.ui.tag.breadcrumbtrail.AbstractNodeDiscovery;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.Util;

import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.PageContext;


/**
 * <p>
 * This tag and implementation of the NodeDiscovery will discover the current node from the HttpServletRequest.
 * </p>
 *
 * <p>
 * This implementation will return a valid uri (with query string if specified) in the getCurrentNode() method.
 * </p>
 *
 * <p>
 * If this is used as a tag in a jsp page, this will define a page variable called
 * BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE with the value a string representing the node uri with query string.
 * </p>
 *
 * <p>
 * <b>Note: </b> This tag/implementation can be used to discover the node based on an HttpServletRequest.
 * This class can be called directly to discover the node or can be used as a tag in the JSP page that will define
 * a page variable containing the node string (which will later be used by the BreadCrumbTrailTag).
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 *
 */
public class HttpServletNodeDiscoveryTag extends AbstractNodeDiscovery {
    /**
     * <p>
     * Constructs the HttpServletNodeDiscoveryTag using no constructor parameters.
     * </p>
     *
     */
    public HttpServletNodeDiscoveryTag() {
        // empty
    }

    /**
     * <p>
     * Discovers the node url from an HttpServletRequest.
     * </p>
     *
     * <p>
     * If the servlet request is not of HttpServletRequest type, a null is returned.
     * </p>
     *
     * <p>
     * Note, the query string will be included as part of the url of the returned BreadCrumbNode instance.
     * </p>
     *
     * @param pageContext A non-null page context to use
     * @return A null if not applicable or a non-null, non-empty string representing the contextual uri
     * (with query string)
     *
     * @throws IllegalArgumentException if the pageContext is null
     * @throws BreadCrumbException when UnsupportedEncodingException occurs
     *
     */
    public BreadCrumbNode getCurrentNode(PageContext pageContext) throws BreadCrumbException {
        Util.checkNull(pageContext, "pageContext");

        ServletRequest request = pageContext.getRequest();

        if (!(request instanceof HttpServletRequest)) {
            return null;
        }

        HttpServletRequest httpServletRequest = (HttpServletRequest) request;

        String url = null;

        // the query string will be included if it is not null or empty.
        if (httpServletRequest.getQueryString() == null || httpServletRequest.getQueryString().trim().length() == 0) {
            url = httpServletRequest.getRequestURI();
        } else {
            url = httpServletRequest.getRequestURI() + "?" + httpServletRequest.getQueryString();
        }

        try {
            return new BreadCrumbNode(URLDecoder.decode(url, "UTF-8"), getTitle());
        } catch (UnsupportedEncodingException e) {
            throw new BreadCrumbException("UnsupportedEncodingException occurs, caused by " + e.getMessage(), e);
        }
    }
}
