/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodediscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;
import com.topcoder.web.ui.tag.breadcrumbtrail.Util;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.BodyTagSupport;


/**
 * <p>
 * This class provides a tag (only) implementation of node discovery.
 * </p>
 *
 * <p>
 * This class will allow the user to hardcode the url into the body of this tag and specify the title in the
 * attributes. That hardcoded information is then read by this tag, combined with the title to make a BreadCrumbNode
 * and that node is put in a page level attribute called BreadCrumbTrailTag.BREADCRUMB_CURRENT_PAGE.
 * </p>
 *
 * <p>
 * <b>Note: </b>
 * This tag will be used by the application to specify the node url via the body of this tag.
 * This can be done by the application where the node url cannot be discovered by any implementations in this
 * component or if the application simply wishes to override or otherwise hard code the url.
 * This class cannot be used automatically by the BreadCrumbTrailTag because it requires user input (and therefore
 * doesn't implement the NodeDiscovery interface).
 * This class simply defines the current page attribute given the body text.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 *
 */
public class InlineTextNodeDiscoveryTag extends BodyTagSupport {
    /**
     * <p>
     * Represents the title associated with this node.
     * </p>
     *
     * <p>
     * This variable is set initially to null, is mutable and can be null but not an empty (trim'd) string.
     * A null represents that the title for the current node should come from the PathDiscovery data source instead.
     * </p>
     *
     * <p>
     * This variable is referenced by the associated getter/setter methods
     * </p>
     */
    private String title;

    /**
     * <p>
     * Constructs the InlineTextNodeDiscoveryTag using no constructor parameters.
     * </p>
     *
     */
    public InlineTextNodeDiscoveryTag() {
        // empty
    }

    /**
     * <p>
     * Starts the processing of this tag by telling the BodyTagSupport class to evaluate the body.
     * </p>
     *
     * @return Always EVAL_BODY_BUFFERED to evaluate the body of this tag
     */
    public int doStartTag() {
        return EVAL_BODY_BUFFERED;
    }

    /**
     * <p>
     * Processes the body content and constructs a BreadCrumbNode according to the information provided by this
     * tag to be stored in BREADCRUMB_CURRENT_NODE attribute under page context for future using.
     * </p>
     *
     * <p>
     * Note, the body is never written out as this tag doesn't render anything.
     * </p>
     *
     * @return Always SKIP_BODY to stop evaluating the body of this tag
     *
     * @throws BreadCrumbException if the body contained nothing
     */
    public int doAfterBody() throws BreadCrumbException {
        String content = bodyContent.getString();

        if (content == null || content.trim().length() == 0) {
            throw new BreadCrumbException("The body content is null or empty.");
        }

        BreadCrumbNode node = new BreadCrumbNode(content, getTitle());

        pageContext.setAttribute(BreadCrumbTrailTag.BREADCRUMB_CURRENT_NODE, node, PageContext.PAGE_SCOPE);

        return SKIP_BODY;
    }

    /**
     * <p>
     * Returns value of title property.
     * </p>
     *
     * @return The possibly null, non-empty string representing the title
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>
     * Sets value of title property.
     * </p>
     *
     * <p>
     * Note, if the title is an empty (trim'd) string, null will be used instead.
     * </p>
     *
     *
     * @param title The possibly null, possibly empty string representing the title
     */
    public void setTitle(String title) {
        this.title = (Util.isEmptyString(title) ? null : title);
    }
}
