/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.nodeformatter;

import java.io.IOException;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.TagSupport;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbTrailTag;
import com.topcoder.web.ui.tag.breadcrumbtrail.Util;

/**
 * <p>
 * This tag will format the node and title using the html anchor pattern.
 * </p>
 *
 * <p>
 * <b>Note: </b> This node formatter tag will be used in the body of a BreadCrumbTrailTag and will
 * format the current node/title using an html anchor.
 * </p>
 *
 * <p>
 * Example output after rendering:
 * <ol>
 * <li>
 * Example when no style class is specified:
 * &lt;a href=&quot;node.getURL()&quot;&gt;title&lt;/a&gt;
 * </li>
 * <li>
 * Example when style class is specified:
 * &lt;a class=&quot;node-style&quot; href=&quot;node.getURL()&quot;&gt;title&lt;/a&gt;
 * </li>
 * </ol>
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 *
 */
public class HtmlAnchorNodeFormatterTag extends TagSupport {
    /**
     * <p>
     * This static constant represents buffer size to be used in StringBuffer construction.
     * </p>
     */
    private static final int BUFFER_SIZE = 255;

    /**
     * <p>
     * This static constant represents the place holder string used in template string(that is, STYLE_TEMPLATE and
     * HREF_TEMPLATE).
     * </p>
     */
    private static final String PLACE_HOLDER = "CONTENT";

    /**
     * <p>
     * This static constant represents the style class template used in html anchor tag.
     * </p>
     */
    private static final String STYLE_TEMPLATE = " class=\"CONTENT\"";

    /**
     * <p>
     * This static constant represents the href template used in html anchor tag.
     * </p>
     */
    private static final String HREF_TEMPLATE = " href=\"CONTENT\"";

    /**
     * <p>
     * This static constant represents the default title when no title is specified.
     * </p>
     */
    private static final String DEFAULT_TITLE = "Unspecified";

    /**
     * <p>
     * Constructs the HtmlAnchorNodeFormatterTag using no constructor parameters.
     * </p>
     *
     */
    public HtmlAnchorNodeFormatterTag() {
        // empty
    }

    /**
     * <p>
     * This method will format the current node and style as a simple html anchor and then write the results out.
     * </p>
     *
     * @return Always SKIP_BODY to skip body evaluation
     *
     * @throws BreadCrumbException if any exception occurs
     */
    public int doStartTag() throws BreadCrumbException {
        // locates the node to be formatted and the style class to be used in page context
        // if the node to be formatted can't be found, then BreadCrumbException will be thrown
        // the style class is allowed to be null or empty
        BreadCrumbNode node = (BreadCrumbNode) Util.getPageContextAttribute(pageContext,
                BreadCrumbTrailTag.BREADCRUMB_FORMAT_NODE, PageContext.PAGE_SCOPE, BreadCrumbNode.class, false);

        String style = (String) Util.getPageContextAttribute(pageContext,
                BreadCrumbTrailTag.BREADCRUMB_FORMAT_STYLE_CLASS, PageContext.PAGE_SCOPE, String.class, true);

        StringBuffer buffer = new StringBuffer(BUFFER_SIZE);
        buffer.append("<a");
        if (style != null && style.trim().length() != 0) {
            buffer.append(STYLE_TEMPLATE.replaceFirst(PLACE_HOLDER, style));
        }
        buffer.append(HREF_TEMPLATE.replaceFirst(PLACE_HOLDER, node.getUrl()));
        buffer.append(">");
        // if the title is null, then default title "Unspecified" will be used
        if (node.getTitle() == null) {
            buffer.append(DEFAULT_TITLE);
        } else {
            buffer.append(node.getTitle());
        }
        buffer.append("</a>");
        try {
            pageContext.getOut().write(buffer.toString());
        } catch (IOException e) {
            throw new BreadCrumbException("IOException occurs when writing a string to jsp page, caused by "
                    + e.getMessage(), e);
        }

        return SKIP_BODY;
    }
}
