/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.servlet.jsp.PageContext;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.MockPageContext;
import com.topcoder.web.ui.tag.breadcrumbtrail.TestHelper;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.nodematchers.RegexNodeMatcher;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.framework.Test;

/**
 * <p>
 * Unit test cases for DynamicPathDiscovery.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 2.0
 */
public class DynamicPathDiscoveryTests extends TestCase {

    /**
     * <p>
     * PageContext instance for testing.
     * </p>
     */
    private PageContext pageContext;

    /**
     * <p>
     * DynamicPathDiscovery instance for testing.
     * </p>
     */
    private DynamicPathDiscovery pathDiscovery;

    /**
     * <p>
     * Setup test environment.
     * </p>
     *
     */
    protected void setUp() {
        pageContext = new MockPageContext();
        pathDiscovery = new DynamicPathDiscovery();
    }

    /**
     * <p>
     * Tears down test environment.
     * </p>
     *
     */
    protected void tearDown() {
        pageContext = null;
        pathDiscovery = null;
    }

    /**
     * <p>
     * Return all tests.
     * </p>
     *
     * @return all tests
     */
    public static Test suite() {
        return new TestSuite(DynamicPathDiscoveryTests.class);
    }

    /**
     * <p>
     * Tests ctor DynamicPathDiscovery#DynamicPathDiscovery() for accuracy.
     * Expects pathDiscovery is successfully created.
     * </p>
     */
    public void testCtor() {
        assertNotNull("Failed to create new DynamicPathDiscovery instance.", pathDiscovery);
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when the current node is a root node and expects correct path returned.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = generatCurrentPath();

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/", "MyMain");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/", "MyMain")};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath2() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = generatCurrentPath();

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/forums", "Forums");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/", "MyMain"),
            new BreadCrumbNode("/forums", "Forums")};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when the current node has a null title and expects the title is replaced on current path.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath3() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = generatCurrentPath();

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/forums?id=888", null);

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/", "MyMain"),
            new BreadCrumbNode("/forums", "Forums"), new BreadCrumbNode("/forums?id=888", "Help")};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * Expects the shortest path should be returned when two paths are available.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath4() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = generatCurrentPathNullNodeMatcher();

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/forums?id=888", null);

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/main", "AltMain"),
            new BreadCrumbNode("/forums?id=888", "Help")};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when more than shortest paths are available and expects the first one is returned.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath5() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = new NodePair[] {
            new NodePair(new BreadCrumbNode("/", "MyMain"), new RegexNodeMatcher("MyMain", "/", "/"))
        };

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/questions", "Questions");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/", "MyMain"),
            new BreadCrumbNode("/topics", "Topics"), node};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath6() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = new NodePair[] {
            new NodePair(new BreadCrumbNode("/main", "AltMain"),
                        new RegexNodeMatcher("AltMain", "/main", "/main"))
        };

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/questions", "Questions");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/main", "AltMain"),
            new BreadCrumbNode("/topics", "Topics"), node};
        assertTrue("Failed to get correct bread crumb trail.",
            TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when the current node is not found in site map.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath7() throws Exception {
        Map siteMap = generateSiteMap();
        NodePair[] currentPath = new NodePair[] {
            new NodePair(new BreadCrumbNode("/main", "AltMain"),
                new RegexNodeMatcher("AltMain", "/main", "/main"))
        };

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_PATH, currentPath, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/unknown", "Unknown");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/main", "AltMain"),
            node};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when the current path is not available.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath8() throws Exception {
        Map siteMap = generateSiteMap();

        pageContext.setAttribute(DynamicPathDiscovery.CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        BreadCrumbNode node = new BreadCrumbNode("/topics", "Topics");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/main", "AltMain"),
            node};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for accuracy.
     * </p>
     *
     * <p>
     * It tests the case when the current path and site map are not available in page context.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath9() throws Exception {
        TestHelper.loadMultipleXMLConfig(DynamicPathDiscovery.class.getName(), "test_files/config/config.xml");

        BreadCrumbNode node = new BreadCrumbNode("/topics", "Topics");

        BreadCrumbNode[] trails = pathDiscovery.getPath(node, pageContext);

        assertNotNull("Expects bread crumb trail found.", trails);

        BreadCrumbNode[] expectedTrails = new BreadCrumbNode[] {new BreadCrumbNode("/main", "AltMain"),
            node};
        assertTrue("Failed to get correct bread crumb trail.",
                TestHelper.isBreadCrumbNodeArrayEquals(trails, expectedTrails));

        TestHelper.clearConfigFile(DynamicPathDiscovery.class.getName());
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for failure.
     * It tests the case that when currentNode is null and expects IllegalArgumentException.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath_NullCurrentNode() throws Exception {
        try {
            pathDiscovery.getPath(null, new MockPageContext());
            fail("IllegalArgumentException expected.");
        } catch (IllegalArgumentException iae) {
            //good
        }
    }

    /**
     * <p>
     * Tests DynamicPathDiscovery#getPath(BreadCrumbNode,PageContext) for failure.
     * It tests the case that when pageContext is null and expects IllegalArgumentException.
     * </p>
     *
     * @throws Exception to JUnit.
     */
    public void testGetPath_NullPageContext() throws Exception {
        try {
            pathDiscovery.getPath(new BreadCrumbNode("/main", "Main"), null);
            fail("IllegalArgumentException expected.");
        } catch (IllegalArgumentException iae) {
            //good
        }
    }

    /**
     * <p>
     * Returns a current path for testing.
     * </p>
     *
     * @return a current path for testing.
     *
     * @throws Exception to JUnit
     */
    private NodePair[] generatCurrentPath() throws Exception {
        NodePair[] currentPath = new NodePair[2];

        currentPath[0] = new NodePair(new BreadCrumbNode("/", "MyMain"),
                new RegexNodeMatcher("MyMain", "/", "/"));
        currentPath[1] = new NodePair(new BreadCrumbNode("/forums", "Forums"),
                new RegexNodeMatcher("Forums", "/forums", "/forums"));

        return currentPath;
    }

    /**
     * <p>
     * Returns a current path with nodeMathcers is null for each node for testing.
     * </p>
     *
     * @return a current path for testing.
     *
     * @throws Exception to JUnit
     */
    private NodePair[] generatCurrentPathNullNodeMatcher() throws Exception {
        NodePair[] currentPath = new NodePair[2];

        currentPath[0] = new NodePair(new BreadCrumbNode("/", "MyMain"), null);
        currentPath[1] = new NodePair(new BreadCrumbNode("/forums", "Forums"), null);

        return currentPath;
    }

    /**
     * <p>
     * Generates a site map for testing.
     * </p>
     *
     * @return a site map for testing.
     *
     * @throws Exception to JUnit
     */
    private Map generateSiteMap() throws Exception {
        NodeMatcher myMainNodeMatcher = new RegexNodeMatcher("MyMain", "/", "/");
        NodeMatcher forumsNodeMatcher = new RegexNodeMatcher("Forums", "/forums", "/forums");
        NodeMatcher helpNodeMatcher = new RegexNodeMatcher("Help", "/forums", "/forums[?]id.*");
        NodeMatcher topicsNodeMatcher = new RegexNodeMatcher("Topics", "/topics", "/topics");

        NodeMatcher questionsNodeMatcher = new RegexNodeMatcher("Questions", "/questions", "/questions");

        NodeMatcher altMainNodeMatcher = new RegexNodeMatcher("AltMain", "/main", "/");

        Map siteMap = new HashMap();

        List myMainChildren = new LinkedList();
        myMainChildren.add(forumsNodeMatcher);
        myMainChildren.add(topicsNodeMatcher);

        List forumsChildren = new LinkedList();
        forumsChildren.add(helpNodeMatcher);

        List altMainChildren = new LinkedList();
        altMainChildren.add(helpNodeMatcher);
        altMainChildren.add(topicsNodeMatcher);

        List topicsChildren = new LinkedList();
        topicsChildren.add(helpNodeMatcher);
        topicsChildren.add(questionsNodeMatcher);

        siteMap.put(myMainNodeMatcher, myMainChildren);
        siteMap.put(forumsNodeMatcher, forumsChildren);
        siteMap.put(helpNodeMatcher, new LinkedList());
        siteMap.put(topicsNodeMatcher, topicsChildren);
        siteMap.put(questionsNodeMatcher, new LinkedList());

        siteMap.put(altMainNodeMatcher, altMainChildren);

        return siteMap;
    }

}